#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>

enum CharacterCategory {
  Body,
  Hair,
  Face,
  Tail,
  Ear,
};
typedef uint8_t CharacterCategory;

enum ColumnDataType {
  String = 0,
  Bool = 1,
  Int8 = 2,
  UInt8 = 3,
  Int16 = 4,
  UInt16 = 5,
  Int32 = 6,
  UInt32 = 7,
  Float32 = 9,
  Int64 = 10,
  UInt64 = 11,
  PackedBool0 = 25,
  PackedBool1 = 26,
  PackedBool2 = 27,
  PackedBool3 = 28,
  PackedBool4 = 29,
  PackedBool5 = 30,
  PackedBool6 = 31,
  PackedBool7 = 32,
};
typedef uint16_t ColumnDataType;

/**
 * Where the existing installation came from
 */
typedef enum ExistingInstallType {
  /**
   * Installed via the official launcher
   */
  OfficialLauncher,
  /**
   * Installed via XIVQuickLauncher
   */
  XIVQuickLauncher,
  /**
   * Installed via XIVLauncherCore
   */
  XIVLauncherCore,
  /**
   * Installed via XIVOnMac
   */
  XIVOnMac,
  /**
   * Installed via Astra
   */
  Astra,
} ExistingInstallType;

/**
 * The playable genders in the game.
 */
enum Gender {
  Male = 0,
  Female = 1,
};
typedef uint8_t Gender;

/**
 * The language the game data is written for. Some of these languages are supported in the Global region.
 */
enum Language {
  /**
   * Used for data that is language-agnostic.
   */
  None,
  /**
   * Japanese language.
   */
  Japanese,
  /**
   * English language.
   */
  English,
  /**
   * German language.
   */
  German,
  /**
   * French language.
   */
  French,
  /**
   * Chinese (Simplified) language.
   */
  ChineseSimplified,
  /**
   * Chinese (Traditional) language.
   */
  ChineseTraditional,
  /**
   * Korean language.
   */
  Korean,
};
typedef uint8_t Language;

/**
 * The kind of patch list.
 * This must be the kind of patch list you're parsing, or else the fields will filled in the wrong order.
 */
typedef enum PatchListType {
  /**
   * A boot patch list.
   */
  Boot,
  /**
   * A game patch ist.
   */
  Game,
} PatchListType;

/**
 * The playable races in the game.
 */
enum Race {
  Hyur = 1,
  Elezen = 2,
  Lalafell = 3,
  Miqote = 4,
  Roegadyn = 5,
  AuRa = 6,
  Hrothgar = 7,
  Viera = 8,
};
typedef uint8_t Race;

typedef enum ShaderStage {
  Vertex,
  Pixel,
} ShaderStage;

/**
 * The slot the item is for.
 */
enum Slot {
  /**
   * The head slot. Shorthand is "met".
   */
  Head,
  /**
   * The hands slot. Shorthand is "glv".
   */
  Hands,
  /**
   * The legs slot. Shorthand is "dwn".
   */
  Legs,
  /**
   * The feet slot. Shorthand is "sho".
   */
  Feet,
  /**
   * The body or chest slot. Shorthand is "top".
   */
  Body,
  /**
   * The earrings slot. Shorthand is "ear".
   */
  Earring,
  /**
   * The neck slot. Shorthand is "nek".
   */
  Neck,
  /**
   * The wrists slot. Shorthand is "wrs".
   */
  Wrists,
  /**
   * The right ring slot. Shorthand is "ril".
   */
  RingLeft,
  /**
   * The left ring slot. Shorthand is "rir".
   */
  RingRight,
};
typedef uint8_t Slot;

typedef enum TextureType {
  TwoDimensional,
  ThreeDimensional,
} TextureType;

/**
 * The playable tribes in the game.
 * Each race has two similar-looking tribes, with the exception of Highlander Hyur which are visually distinct.
 */
enum Tribe {
  Midlander = 1,
  Highlander = 2,
  Wildwood = 3,
  Duskwight = 4,
  Plainsfolk = 5,
  Dunesfolk = 6,
  Seeker = 7,
  Keeper = 8,
  SeaWolf = 9,
  Hellsguard = 10,
  Raen = 11,
  Xaela = 12,
  Hellion = 13,
  Lost = 14,
  Rava = 15,
  Veena = 16,
};
typedef uint8_t Tribe;

/**
 * The format of the vertex stream.
 */
enum VertexType {
  /**
   * 1 32-bit float
   */
  Single1 = 0,
  /**
   * 2 32-bit floats
   */
  Single2 = 1,
  /**
   * 3 32-bit floats
   */
  Single3 = 2,
  /**
   * 4 32-bit floats
   */
  Single4 = 3,
  /**
   * 4 bytes
   */
  Byte4 = 5,
  /**
   * 2 16-bit signed integers
   */
  Short2 = 6,
  /**
   * 4 16-bit signed integers
   */
  Short4 = 7,
  /**
   * 4 8-bit floats
   */
  ByteFloat4 = 8,
  /**
   * Duplicate of Short2?
   */
  Short2n = 9,
  /**
   * Duplicate of Short4?
   */
  Short4n = 10,
  /**
   * 2 16-bit floats
   */
  Half2 = 13,
  /**
   * 4 16-bit floats
   */
  Half4 = 14,
  /**
   * 2 16-bit unsigned integers
   */
  UnsignedShort2 = 16,
  /**
   * 4 16-bit unsigned integers
   */
  UnsignedShort4 = 17,
};
typedef uint8_t VertexType;

/**
 * What the vertex stream is used for.
 */
enum VertexUsage {
  Position = 0,
  BlendWeights = 1,
  BlendIndices = 2,
  Normal = 3,
  UV = 4,
  Tangent = 5,
  BiTangent = 6,
  Color = 7,
};
typedef uint8_t VertexUsage;

/**
 * Implementation of the Blowfish block cipher, specialized for encrypting and decrypting SqexArg - the technique used to encrypt game arguments by the launcher.
 *
 * # Example
 *
 * ```
 * # use physis::blowfish::Blowfish;
 * let key = b"abcdefgh";
 * let data = b"foobar  ";
 *
 * let fish = Blowfish::new(key);
 * let encrypted = fish.encrypt(data).unwrap();
 * let decrypted = fish.decrypt(&encrypted).unwrap();
 * # assert_eq!(data, &decrypted[..])
 * ```
 */
typedef struct Blowfish Blowfish;

/**
 * Represents the boot data for FFXIV, which is located under the "boot" directory.
 */
typedef struct BootData BootData;

typedef struct CMP CMP;

/**
 * Represents a config file, which is made up of categories and settings. Categories may have zero to one setting.
 */
typedef struct ConfigFile ConfigFile;

/**
 * An Excel data file. Represents a page in an Excel Sheet.
 */
typedef struct EXD EXD;

typedef struct EXH EXH;

/**
 * Framework for operating on game data.
 */
typedef struct GameData GameData;

typedef struct MDL MDL;

typedef struct PreBoneDeformer PreBoneDeformer;

typedef struct Schd Schd;

typedef struct ShaderPackage ShaderPackage;

/**
 * Implementation of the Blowfish block cipher, specialized for encrypting and decrypting Steam tickets in the launcher. This is somehow *different* than the one used with SqexArg.
 */
typedef struct SteamTicketBlowfish SteamTicketBlowfish;

typedef struct physis_Buffer {
  uint32_t size;
  uint8_t *data;
} physis_Buffer;

/**
 * The type of repository, discerning game data from expansion data.
 */
typedef enum RepositoryType_Tag {
  /**
   * The base game directory, like "ffxiv".
   */
  Base,
  /**
   * An expansion directory, like "ex1".
   */
  Expansion,
} RepositoryType_Tag;

typedef struct Expansion_Body {
  /**
   * The expansion number starting at 1.
   */
  int32_t number;
} Expansion_Body;

typedef struct RepositoryType {
  RepositoryType_Tag tag;
  union {
    Expansion_Body expansion;
  };
} RepositoryType;

typedef struct physis_Repository {
  const char *name;
  struct RepositoryType repository_type;
  const char *version;
} physis_Repository;

typedef struct physis_Repositories {
  uint32_t repositories_count;
  struct physis_Repository *repositories;
} physis_Repositories;

typedef struct physis_EXH {
  struct EXH *p_ptr;
  uint32_t page_count;
  uint32_t language_count;
  Language *languages;
  uint32_t column_count;
  uint32_t row_count;
  ColumnDataType *column_types;
} physis_EXH;

typedef struct physis_EXD {
  struct EXD *p_ptr;
  unsigned int column_count;
  unsigned int row_count;
} physis_EXD;

typedef struct physis_SheetNames {
  uint32_t name_count;
  const char **names;
} physis_SheetNames;

/**
 * Represents an element within a bigger vertex stream.
 */
typedef struct VertexElement {
  uint8_t stream;
  uint8_t offset;
  VertexType vertex_type;
  VertexUsage vertex_usage;
  uint8_t usage_index;
} VertexElement;

typedef struct Vertex {
  float position[3];
  float uv0[2];
  float uv1[2];
  float normal[3];
  float bitangent[4];
  float color[4];
  float bone_weight[4];
  uint8_t bone_id[4];
} Vertex;

typedef struct SubMesh {
  uintptr_t submesh_index;
  uint32_t index_count;
  uint32_t index_offset;
} SubMesh;

typedef struct physis_Shape {
  const char *name;
  struct Vertex *morphed_vertices;
} physis_Shape;

typedef struct physis_Part {
  uint32_t num_vertices;
  struct Vertex *vertices;
  const uint8_t **streams;
  uintptr_t *stream_sizes;
  uintptr_t *stream_strides;
  uintptr_t num_streams;
  uint32_t num_indices;
  uint16_t *indices;
  uint16_t material_index;
  uint32_t num_submeshes;
  struct SubMesh *submeshes;
  uint32_t num_shapes;
  struct physis_Shape *shapes;
} physis_Part;

typedef struct physis_LOD {
  uint32_t num_vertex_elements;
  struct VertexElement *vertex_elements;
  uint32_t num_parts;
  struct physis_Part *parts;
} physis_LOD;

typedef struct physis_MDL {
  struct MDL *p_ptr;
  uint32_t num_lod;
  struct physis_LOD *lods;
  uint32_t num_affected_bones;
  const char **affected_bone_names;
  uint32_t num_material_names;
  const char **material_names;
} physis_MDL;

typedef struct NewShapeValue {
  uint32_t base_index;
  struct Vertex replacing_vertex;
} NewShapeValue;

typedef struct physis_SupportedTribes {
  Tribe subraces[2];
} physis_SupportedTribes;

typedef struct physis_EXL {
  int32_t version;
  int32_t entry_count;
  const char *const *entry_keys;
  const int32_t *entry_values;
} physis_EXL;

typedef enum physis_ColumnData_Tag {
  String,
  Bool,
  Int8,
  UInt8,
  Int16,
  UInt16,
  Int32,
  UInt32,
  Float32,
  Int64,
  UInt64,
} physis_ColumnData_Tag;

typedef struct physis_ColumnData {
  physis_ColumnData_Tag tag;
  union {
    struct {
      const char *string;
    };
    struct {
      bool bool_;
    };
    struct {
      int8_t int8;
    };
    struct {
      uint8_t u_int8;
    };
    struct {
      int16_t int16;
    };
    struct {
      uint16_t u_int16;
    };
    struct {
      int32_t int32;
    };
    struct {
      uint32_t u_int32;
    };
    struct {
      float float32;
    };
    struct {
      int64_t int64;
    };
    struct {
      uint64_t u_int64;
    };
  };
} physis_ColumnData;

typedef struct physis_ExcelRow {
  struct physis_ColumnData *column_data;
} physis_ExcelRow;

typedef struct physis_ExcelRows {
  struct physis_ExcelRow *row_data;
  unsigned int row_count;
} physis_ExcelRows;

typedef struct physis_Bone {
  uint32_t index;
  const char *name;
  struct physis_Bone *parent_bone;
  uint32_t parent_index;
  float position[3];
  float rotation[4];
  float scale[3];
} physis_Bone;

typedef struct physis_Skeleton {
  uint32_t num_bones;
  struct physis_Bone *bones;
  struct physis_Bone *root_bone;
} physis_Skeleton;

typedef struct physis_Texture {
  enum TextureType texture_type;
  uint32_t width;
  uint32_t height;
  uint32_t depth;
  uint32_t rgba_size;
  uint8_t *rgba;
} physis_Texture;

typedef struct ShaderKey {
  uint32_t category;
  uint32_t value;
} ShaderKey;

typedef struct Constant {
  uint32_t id;
  uint32_t num_values;
  float values[4];
} Constant;

typedef struct Sampler {
  /**
   * This is a CRC hash, it can be calculated via ShaderPackage::crc
   */
  uint32_t texture_usage;
  uint32_t flags;
  uint8_t texture_index;
  uint8_t unknown1;
  uint8_t unknown2;
  uint8_t unknown3;
} Sampler;

typedef struct LegacyColorTableRow {
  float diffuse_color[3];
  float specular_strength;
  float specular_color[3];
  float gloss_strength;
  float emissive_color[3];
  uint16_t tile_set;
  float material_repeat[2];
  float material_skew[2];
} LegacyColorTableRow;

typedef struct physis_LegacyColorTable {
  uint32_t num_rows;
  struct LegacyColorTableRow *rows;
} physis_LegacyColorTable;

typedef struct DawntrailColorTableRow {
  float diffuse_color[3];
  float unknown1;
  float specular_color[3];
  float unknown2;
  float emissive_color[3];
  float unknown3;
  float sheen_rate;
  float sheen_tint;
  float sheen_aperture;
  float unknown4;
  float roughness;
  float unknown5;
  float metalness;
  float anisotropy;
  float unknown6;
  float sphere_mask;
  float unknown7;
  float unknown8;
  uint16_t shader_index;
  uint16_t tile_set;
  float tile_alpha;
  uint16_t sphere_index;
  float material_repeat[2];
  float material_skew[2];
} DawntrailColorTableRow;

typedef struct physis_DawntrailColorTable {
  uint32_t num_rows;
  struct DawntrailColorTableRow *rows;
} physis_DawntrailColorTable;

typedef struct physis_Material {
  const char *shpk_name;
  uint32_t num_textures;
  const char **textures;
  uint32_t num_shader_keys;
  struct ShaderKey *shader_keys;
  uint32_t num_constants;
  struct Constant *constants;
  uint32_t num_samplers;
  struct Sampler *samplers;
  struct physis_LegacyColorTable legacy_color_table;
  struct physis_DawntrailColorTable dawntrail_color_table;
} physis_Material;

typedef struct physis_ShaderParameter {
  uint16_t slot;
  const char *name;
} physis_ShaderParameter;

typedef struct physis_Shader {
  int32_t len;
  uint8_t *bytecode;
  int32_t num_scalar_parameters;
  struct physis_ShaderParameter *scalar_parameters;
  int32_t num_resource_parameters;
  struct physis_ShaderParameter *resource_parameters;
} physis_Shader;

typedef struct Key {
  uint32_t id;
  uint32_t default_value;
} Key;

typedef struct MaterialParameter {
  uint32_t id;
  uint16_t byte_offset;
  uint16_t byte_size;
} MaterialParameter;

typedef struct physis_SHPK {
  struct ShaderPackage *p_ptr;
  int32_t num_vertex_shaders;
  struct physis_Shader *vertex_shaders;
  int32_t num_pixel_shaders;
  struct physis_Shader *pixel_shaders;
  int32_t num_system_keys;
  struct Key *system_keys;
  int32_t num_scene_keys;
  struct Key *scene_keys;
  int32_t num_material_keys;
  struct Key *material_keys;
  uint32_t sub_view_key1_default;
  uint32_t sub_view_key2_default;
  uint32_t material_parameters_size;
  uint32_t num_material_parameters;
  struct MaterialParameter *material_parameters;
} physis_SHPK;

typedef struct Pass {
  uint32_t id;
  uint32_t vertex_shader;
  uint32_t pixel_shader;
} Pass;

typedef struct physis_SHPKNode {
  uint32_t selector;
  uint32_t pass_count;
  uint8_t pass_indices[16];
  uint32_t system_key_count;
  uint32_t *system_keys;
  uint32_t scene_key_count;
  uint32_t *scene_keys;
  uint32_t material_key_count;
  uint32_t *material_keys;
  uint32_t subview_key_count;
  uint32_t *subview_keys;
  struct Pass *passes;
} physis_SHPKNode;

typedef struct physis_CMP {
  struct CMP *p_ptr;
} physis_CMP;

typedef struct RacialScalingParameters {
  /**
   * Male minimum height
   */
  float male_min_size;
  /**
   * Male maximum height
   */
  float male_max_size;
  /**
   * Male minimum tail size
   */
  float male_min_tail;
  /**
   * Male maximum tail size
   */
  float male_max_tail;
  /**
   * Female minimum height
   */
  float female_min_size;
  /**
   * Female maximum height
   */
  float female_max_size;
  /**
   * Female minimum tail size
   */
  float female_min_tail;
  /**
   * Female maximum tail size
   */
  float female_max_tail;
  /**
   * Minimum bust size on the X-axis
   */
  float bust_min_x;
  /**
   * Minimum bust size on the Y-axis
   */
  float bust_min_y;
  /**
   * Minimum bust size on the Z-axis
   */
  float bust_min_z;
  /**
   * Maximum bust size on the X-axis
   */
  float bust_max_x;
  /**
   * Maximum bust size on the Y-axis
   */
  float bust_max_y;
  /**
   * Maximum bust size on the Z-axis
   */
  float bust_max_z;
} RacialScalingParameters;

typedef struct CustomizeData {
  /**
   * The race of the character.
   */
  Race race;
  /**
   * The gender of the character.
   */
  Gender gender;
  /**
   * The age of the character. Normal = 1, Old = 3, Young = 4.
   */
  uint8_t age;
  /**
   * The height of the character from 0 to 100.
   */
  uint8_t height;
  /**
   * The character's tribe.
   */
  Tribe tribe;
  /**
   * The character's selected face.
   */
  uint8_t face;
  /**
   * The character's selected hair.
   */
  uint8_t hair;
  /**
   * If hair highlights are enabled for this character.
   */
  bool enable_highlights;
  /**
   * The character's skin tone.
   */
  uint8_t skin_tone;
  /**
   * The character's right eye color.
   */
  uint8_t right_eye_color;
  /**
   * The character's hair color.
   */
  uint8_t hair_tone;
  /**
   * The color of the hair highlights.
   */
  uint8_t highlights;
  /**
   * The selected facial features.
   */
  uint8_t facial_features;
  /**
   * The color of the character's facial features.'
   */
  uint8_t facial_feature_color;
  /**
   * The character's selected eyebrows.
   */
  uint8_t eyebrows;
  /**
   * The character's left eye color.
   */
  uint8_t left_eye_color;
  /**
   * The character's selected eyes.
   * If the character has selected "Small Iris" then it adds 128 to this.
   */
  uint8_t eyes;
  /**
   * The character's selected nose.
   */
  uint8_t nose;
  /**
   * The character's selected jaw.
   */
  uint8_t jaw;
  /**
   * The character's selected mouth.
   */
  uint8_t mouth;
  /**
   * The character's selected pattern.
   */
  uint8_t lips_tone_fur_pattern;
  /**
   * Depending on the race, it's either the ear size, muscle size, or tail size.
   */
  uint8_t race_feature_size;
  /**
   * Depending on the race, it's the selected tail or ears.
   */
  uint8_t race_feature_type;
  /**
   * The size of the character's bust from 0 to 100.
   */
  uint8_t bust;
  /**
   * The character's choice of face paint.
   */
  uint8_t face_paint;
  /**
   * The color of the face paint.
   * If the character has selected "Light" then it adds 128 to this.
   */
  uint8_t face_paint_color;
  /**
   * The character's chosen voice.
   */
  uint8_t voice;
} CustomizeData;

typedef struct physis_CharacterData {
  uint32_t version;
  struct CustomizeData customize;
  uint32_t timestamp;
  const char *comment;
} physis_CharacterData;

typedef struct physis_ConfigFile {
  struct ConfigFile *p_ptr;
} physis_ConfigFile;

typedef struct physis_PBD {
  struct PreBoneDeformer *p_ptr;
} physis_PBD;

typedef struct physis_PreBoneDeformBone {
  const char *name;
  float deform[12];
} physis_PreBoneDeformBone;

typedef struct physis_PreBoneDeformMatrices {
  int32_t num_bones;
  struct physis_PreBoneDeformBone *bones;
} physis_PreBoneDeformMatrices;

typedef struct physis_PlateModel {
  float position[2];
  const char *filename;
} physis_PlateModel;

typedef struct physis_Terrain {
  int32_t num_plates;
  struct physis_PlateModel *plates;
} physis_Terrain;

typedef struct physis_Dictionary {
  int32_t num_words;
  const char *const *words;
} physis_Dictionary;

typedef struct physis_IndexEntries {
  uint32_t num_entries;
  const uint32_t *dir_entries;
  const uint32_t *filename_entries;
} physis_IndexEntries;

/**
 * An existing install location on disk
 */
typedef struct physis_ExistingGameDirectory {
  /**
   * The application where this installation was from
   */
  enum ExistingInstallType install_type;
  /**
   * The path to the "main folder" where "game" and "boot" sits
   */
  const char *path;
  /**
   * The latest game/expansion version in this directory
   */
  const char *version;
} physis_ExistingGameDirectory;

typedef struct physis_ExistingGameDirectories {
  uint32_t count;
  struct physis_ExistingGameDirectory *entries;
} physis_ExistingGameDirectories;

/**
 * An existing user directory
 */
typedef struct physis_ExistingUserDirectory {
  /**
   * The application where this directory was from
   */
  enum ExistingInstallType install_type;
  /**
   * The path to the user folder
   */
  const char *path;
} physis_ExistingUserDirectory;

typedef struct physis_ExistingUserDirectories {
  uint32_t count;
  struct physis_ExistingUserDirectory *entries;
} physis_ExistingUserDirectories;

typedef struct physis_SCHD {
  struct Schd *p_ptr;
  enum ShaderStage shader_stage;
} physis_SCHD;

typedef struct physis_PatchEntry {
  const char *url;
  const char *version;
  uint64_t hash_count;
  const char *const *hashes;
  int64_t hash_block_size;
  int64_t length;
  int64_t size_on_disk;
} physis_PatchEntry;

typedef struct physis_PatchList {
  uint64_t patch_length;
  int32_t num_entries;
  const struct physis_PatchEntry *entries;
  int64_t total_size_downloaded;
} physis_PatchList;

typedef struct Transformation {
  float translation[3];
  float rotation[3];
  float scale[3];
} Transformation;

typedef struct physis_BGInstanceObject {
  const char *asset_path;
} physis_BGInstanceObject;

typedef enum physis_LayerEntry_Tag {
  None,
  BG,
} physis_LayerEntry_Tag;

typedef struct physis_LayerEntry {
  physis_LayerEntry_Tag tag;
  union {
    struct {
      struct physis_BGInstanceObject bg;
    };
  };
} physis_LayerEntry;

typedef struct physis_InstanceObject {
  uint32_t instance_id;
  const char *name;
  struct Transformation transform;
  struct physis_LayerEntry data;
} physis_InstanceObject;

typedef struct physis_Layer {
  const struct physis_InstanceObject *objects;
  uint32_t num_objects;
  const char *name;
  uint32_t id;
} physis_Layer;

typedef struct physis_LayerChunk {
  const struct physis_Layer *layers;
  uint32_t num_layers;
} physis_LayerChunk;

typedef struct physis_LayerGroup {
  const struct physis_LayerChunk *chunks;
  uint32_t num_chunks;
} physis_LayerGroup;

const char *physis_get_physis_version(void);

const char *physis_get_libphysis_version(void);

struct physis_Buffer physis_read_file(const char *path);

void physis_free_file(const struct physis_Buffer *buffer);

/**
 * Checks if the file at `path` exists.
 */
bool physis_gamedata_exists(struct GameData *game_data, const char *path);

void physis_gamedata_free_repositories(struct physis_Repositories repositories);

/**
 * Extracts the raw game file from `path`, and puts it in `data` with `size` length. If the path was not found,
 * `size` is 0 and `data` is NULL.
 */
struct physis_Buffer physis_gamedata_extract_file(struct GameData *game_data,
                                                  const char *path);

uint64_t physis_gamedata_find_offset(struct GameData *game_data, const char *path);

void physis_gamedata_free_sheet_header(struct physis_EXH*);

/**
 * Initializes a new GameData structure.
 */
struct GameData *physis_gamedata_initialize(const char *path);

void physis_gamedata_free(struct GameData *game_data);

struct physis_Repositories physis_gamedata_get_repositories(const struct GameData *game_data);

struct physis_EXD physis_gamedata_read_excel_sheet(struct GameData *game_data,
                                                   const char *name,
                                                   const struct physis_EXH *exh,
                                                   Language language,
                                                   unsigned int page);

const char *physis_gamedata_get_exd_filename(const char *name,
                                             const struct physis_EXH *exh,
                                             Language language,
                                             unsigned int page);

void physis_gamedata_free_sheet(struct physis_EXD exd);

bool physis_gamedata_apply_patch(const struct GameData *gamedata, const char *path);

struct physis_SheetNames physis_gamedata_get_all_sheet_names(struct GameData *game_data);

const char *physis_bootdata_get_version(const struct BootData *boot_data);

/**
 * Initializes a new BootData structure.
 */
struct BootData *physis_bootdata_initialize(const char *path);

void physis_bootdata_free(struct BootData *boot_data);

bool physis_bootdata_apply_patch(const struct BootData *bootdata, const char *path);

struct physis_MDL physis_mdl_parse(struct physis_Buffer buffer);

struct physis_Buffer physis_mdl_write(const struct physis_MDL *mdl);

void physis_mdl_replace_vertices(struct physis_MDL *mdl,
                                 uint32_t lod_index,
                                 uint32_t part_index,
                                 uint32_t num_vertices,
                                 const struct Vertex *vertices_ptr,
                                 uint32_t num_indices,
                                 const uint16_t *indices_ptr,
                                 uint32_t num_submeshes,
                                 const struct SubMesh *submeshes_ptr);

void physis_mdl_remove_shape_meshes(struct physis_MDL *mdl);

void physis_mdl_add_shape_mesh(struct physis_MDL *mdl,
                               uint32_t lod_index,
                               uint32_t shape_index,
                               uint32_t shape_mesh_index,
                               uint32_t part_index,
                               uint32_t num_shape_values,
                               const struct NewShapeValue *shape_values);

uintptr_t physis_get_vertex_type_size(VertexType vertex_type);

void physis_mdl_free(const struct physis_MDL *mdl);

int32_t physis_get_race_code(Race race, Tribe tribe, Gender gender);

struct physis_SupportedTribes physis_get_supported_tribes(Race race);

struct physis_EXL physis_gamedata_read_excel_list(struct physis_Buffer buffer);

Slot physis_slot_from_id(int32_t slot_id);

const char *physis_get_slot_name(Slot slot);

const char *physis_build_equipment_path(int32_t model_id,
                                        Race race,
                                        Tribe tribe,
                                        Gender gender,
                                        Slot slot);

const char *physis_build_character_path(CharacterCategory category,
                                        int32_t body_ver,
                                        Race race,
                                        Tribe tribe,
                                        Gender gender);

const char *physis_build_skin_material_path(int32_t race_code,
                                            int32_t body_code,
                                            const char *material_name);

const char *physis_build_gear_material_path(int32_t gear_id,
                                            int32_t gear_version,
                                            const char *material_name);

const char *physis_build_face_material_path(int32_t race_code,
                                            int32_t face_code,
                                            const char *material_name);

const char *physis_build_hair_material_path(int32_t race_code,
                                            int32_t hair_code,
                                            const char *material_name);

const char *physis_build_ear_material_path(int32_t race_code,
                                           int32_t ear_code,
                                           const char *material_name);

const char *physis_build_tail_material_path(int32_t race_code,
                                            int32_t tail_code,
                                            const char *material_name);

struct Blowfish *physis_blowfish_initialize(uint8_t *key, unsigned int key_size);

void physis_blowfish_free(struct Blowfish *blowfish);

bool physis_blowfish_encrypt(const struct Blowfish *blowfish,
                             uint8_t *in_data,
                             unsigned int in_data_size,
                             uint8_t **out_data,
                             uint32_t *out_data_size);

bool physis_blowfish_decrypt(const struct Blowfish *blowfish,
                             uint8_t *in_data,
                             unsigned int in_data_size,
                             uint8_t **out_data,
                             uint32_t *out_data_size);

struct SteamTicketBlowfish *miscel_steamticket_blowfish_initialize(uint8_t *key,
                                                                   unsigned int key_size);

void miscel_steamticket_physis_blowfish_free(struct SteamTicketBlowfish *blowfish);

void miscel_steamticket_blowfish_encrypt(const struct SteamTicketBlowfish *blowfish,
                                         uint8_t *in_data,
                                         unsigned int in_data_size);

struct physis_EXH *physis_parse_excel_sheet_header(struct physis_Buffer buffer);

struct physis_ExcelRows physis_exd_read_row(const struct physis_EXD *exd, uint32_t id);

void physis_exd_free_rows(const struct physis_EXD *exd, const struct physis_ExcelRows *rows);

struct physis_Skeleton physis_parse_skeleton(struct physis_Buffer buffer);

struct physis_Texture physis_texture_parse(struct physis_Buffer buffer);

struct physis_Material physis_material_parse(struct physis_Buffer buffer);

struct physis_SHPK physis_parse_shpk(struct physis_Buffer buffer);

struct physis_SHPKNode physis_shpk_get_node(const struct physis_SHPK *shpk, uint32_t key);

uint32_t physis_shpk_build_selector_from_all_keys(const uint32_t *system_keys,
                                                  uint32_t system_key_count,
                                                  const uint32_t *scene_keys,
                                                  uint32_t scene_key_count,
                                                  const uint32_t *material_keys,
                                                  uint32_t material_key_count,
                                                  const uint32_t *subview_keys,
                                                  uint32_t subview_key_count);

uint32_t physis_shpk_crc(const char *name);

struct physis_CMP physis_cmp_parse(struct physis_Buffer buffer);

struct RacialScalingParameters physis_cmp_get_racial_scaling_parameters(struct physis_CMP cmp,
                                                                        Race,
                                                                        Tribe tribe);

struct physis_CharacterData physis_chardat_parse(struct physis_Buffer buffer);

struct physis_ConfigFile physis_cfg_parse(struct physis_Buffer buffer);

void physis_cfg_set_value(struct physis_ConfigFile cfg, const char *key, const char *value);

struct physis_Buffer physis_cfg_write(struct physis_ConfigFile cfg);

struct physis_PBD physis_parse_pbd(struct physis_Buffer buffer);

struct physis_PreBoneDeformMatrices physis_pbd_get_deform_matrix(struct physis_PBD pbd,
                                                                 uint16_t from_body_id,
                                                                 uint16_t to_body_id);

struct physis_Terrain physis_parse_tera(struct physis_Buffer buffer);

struct physis_Dictionary physis_parse_dictionary(struct physis_Buffer buffer);

struct physis_IndexEntries physis_index_parse(const char *path);

uint32_t physis_generate_partial_hash(const char *name);

uint64_t physis_calculate_hash(const char *index_file_path, const char *path);

struct physis_ExistingGameDirectories physis_find_existing_game_dirs(void);

struct physis_ExistingUserDirectories physis_find_existing_user_dirs(void);

struct physis_SCHD physis_parse_schd(struct physis_Buffer _buffer);

const char *physis_extract_frontier_url(const char *launcher_path);

struct physis_PatchList physis_parse_patchlist(enum PatchListType patch_type, const char *encoded);

struct physis_LayerGroup physis_layergroup_read(struct physis_Buffer buffer);
