#pragma once

#include <cstdarg>
#include <cstdint>
#include <cstdlib>

enum class CharacterCategory : uint8_t {
  Body,
  Hair,
  Face,
  Tail,
  Ear,
};

enum class ColumnDataType : uint16_t {
  String = 0,
  Bool = 1,
  Int8 = 2,
  UInt8 = 3,
  Int16 = 4,
  UInt16 = 5,
  Int32 = 6,
  UInt32 = 7,
  Float32 = 9,
  Int64 = 10,
  UInt64 = 11,
  PackedBool0 = 25,
  PackedBool1 = 26,
  PackedBool2 = 27,
  PackedBool3 = 28,
  PackedBool4 = 29,
  PackedBool5 = 30,
  PackedBool6 = 31,
  PackedBool7 = 32,
};

/// Where the existing installation came from
enum class ExistingInstallType {
  /// Installed via the official launcher
  OfficialLauncher,
  /// Installed via XIVQuickLauncher
  XIVQuickLauncher,
  /// Installed via XIVLauncherCore
  XIVLauncherCore,
  /// Installed via XIVOnMac
  XIVOnMac,
  /// Installed via Astra
  Astra,
};

/// The playable genders in the game.
enum class Gender : uint8_t {
  Male = 0,
  Female = 1,
};

/// The language the game data is written for. Some of these languages are supported in the Global region.
enum class Language : uint8_t {
  /// Used for data that is language-agnostic.
  None,
  /// Japanese language.
  Japanese,
  /// English language.
  English,
  /// German language.
  German,
  /// French language.
  French,
  /// Chinese (Simplified) language.
  ChineseSimplified,
  /// Chinese (Traditional) language.
  ChineseTraditional,
  /// Korean language.
  Korean,
};

/// The kind of patch list.
/// This must be the kind of patch list you're parsing, or else the fields will filled in the wrong order.
enum class PatchListType {
  /// A boot patch list.
  Boot,
  /// A game patch ist.
  Game,
};

/// The playable races in the game.
enum class Race : uint8_t {
  Hyur = 1,
  Elezen = 2,
  Lalafell = 3,
  Miqote = 4,
  Roegadyn = 5,
  AuRa = 6,
  Hrothgar = 7,
  Viera = 8,
};

enum class ShaderStage {
  Vertex,
  Pixel,
};

/// The slot the item is for.
enum class Slot : uint8_t {
  /// The head slot. Shorthand is "met".
  Head,
  /// The hands slot. Shorthand is "glv".
  Hands,
  /// The legs slot. Shorthand is "dwn".
  Legs,
  /// The feet slot. Shorthand is "sho".
  Feet,
  /// The body or chest slot. Shorthand is "top".
  Body,
  /// The earrings slot. Shorthand is "ear".
  Earring,
  /// The neck slot. Shorthand is "nek".
  Neck,
  /// The wrists slot. Shorthand is "wrs".
  Wrists,
  /// The right ring slot. Shorthand is "ril".
  RingLeft,
  /// The left ring slot. Shorthand is "rir".
  RingRight,
};

enum class TextureType {
  TwoDimensional,
  ThreeDimensional,
};

/// The playable tribes in the game.
/// Each race has two similar-looking tribes, with the exception of Highlander Hyur which are visually distinct.
enum class Tribe : uint8_t {
  Midlander = 1,
  Highlander = 2,
  Wildwood = 3,
  Duskwight = 4,
  Plainsfolk = 5,
  Dunesfolk = 6,
  Seeker = 7,
  Keeper = 8,
  SeaWolf = 9,
  Hellsguard = 10,
  Raen = 11,
  Xaela = 12,
  Hellion = 13,
  Lost = 14,
  Rava = 15,
  Veena = 16,
};

/// The format of the vertex stream.
enum class VertexType : uint8_t {
  /// 1 32-bit float
  Single1 = 0,
  /// 2 32-bit floats
  Single2 = 1,
  /// 3 32-bit floats
  Single3 = 2,
  /// 4 32-bit floats
  Single4 = 3,
  /// 4 bytes
  Byte4 = 5,
  /// 2 16-bit signed integers
  Short2 = 6,
  /// 4 16-bit signed integers
  Short4 = 7,
  /// 4 8-bit floats
  ByteFloat4 = 8,
  /// Duplicate of Short2?
  Short2n = 9,
  /// Duplicate of Short4?
  Short4n = 10,
  /// 2 16-bit floats
  Half2 = 13,
  /// 4 16-bit floats
  Half4 = 14,
  /// 2 16-bit unsigned integers
  UnsignedShort2 = 16,
  /// 4 16-bit unsigned integers
  UnsignedShort4 = 17,
};

/// What the vertex stream is used for.
enum class VertexUsage : uint8_t {
  Position = 0,
  BlendWeights = 1,
  BlendIndices = 2,
  Normal = 3,
  UV = 4,
  Tangent = 5,
  BiTangent = 6,
  Color = 7,
};

/// Implementation of the Blowfish block cipher, specialized for encrypting and decrypting SqexArg - the technique used to encrypt game arguments by the launcher.
///
/// # Example
///
/// ```
/// # use physis::blowfish::Blowfish;
/// let key = b"abcdefgh";
/// let data = b"foobar  ";
///
/// let fish = Blowfish::new(key);
/// let encrypted = fish.encrypt(data).unwrap();
/// let decrypted = fish.decrypt(&encrypted).unwrap();
/// # assert_eq!(data, &decrypted[..])
/// ```
struct Blowfish;

/// Represents the boot data for FFXIV, which is located under the "boot" directory.
struct BootData;

struct CMP;

/// Represents a config file, which is made up of categories and settings. Categories may have zero to one setting.
struct ConfigFile;

/// An Excel data file. Represents a page in an Excel Sheet.
struct EXD;

struct EXH;

/// Framework for operating on game data.
struct GameData;

struct MDL;

struct PreBoneDeformer;

struct Schd;

struct ShaderPackage;

/// Implementation of the Blowfish block cipher, specialized for encrypting and decrypting Steam tickets in the launcher. This is somehow *different* than the one used with SqexArg.
struct SteamTicketBlowfish;

struct physis_Buffer {
  uint32_t size;
  uint8_t *data;
};

/// The type of repository, discerning game data from expansion data.
struct RepositoryType {
  enum class Tag {
    /// The base game directory, like "ffxiv".
    Base,
    /// An expansion directory, like "ex1".
    Expansion,
  };

  struct Expansion_Body {
    /// The expansion number starting at 1.
    int32_t number;
  };

  Tag tag;
  union {
    Expansion_Body expansion;
  };
};

struct physis_Repository {
  const char *name;
  RepositoryType repository_type;
  const char *version;
};

struct physis_Repositories {
  uint32_t repositories_count;
  physis_Repository *repositories;
};

struct physis_EXH {
  EXH *p_ptr;
  uint32_t page_count;
  uint32_t language_count;
  Language *languages;
  uint32_t column_count;
  uint32_t row_count;
  ColumnDataType *column_types;
};

struct physis_EXD {
  EXD *p_ptr;
  unsigned int column_count;
  unsigned int row_count;
};

struct physis_SheetNames {
  uint32_t name_count;
  const char **names;
};

/// Represents an element within a bigger vertex stream.
struct VertexElement {
  uint8_t stream;
  uint8_t offset;
  VertexType vertex_type;
  VertexUsage vertex_usage;
  uint8_t usage_index;
};

struct Vertex {
  float position[3];
  float uv0[2];
  float uv1[2];
  float normal[3];
  float bitangent[4];
  float color[4];
  float bone_weight[4];
  uint8_t bone_id[4];
};

struct SubMesh {
  uintptr_t submesh_index;
  uint32_t index_count;
  uint32_t index_offset;
};

struct physis_Shape {
  const char *name;
  Vertex *morphed_vertices;
};

struct physis_Part {
  uint32_t num_vertices;
  Vertex *vertices;
  const uint8_t **streams;
  uintptr_t *stream_sizes;
  uintptr_t *stream_strides;
  uintptr_t num_streams;
  uint32_t num_indices;
  uint16_t *indices;
  uint16_t material_index;
  uint32_t num_submeshes;
  SubMesh *submeshes;
  uint32_t num_shapes;
  physis_Shape *shapes;
};

struct physis_LOD {
  uint32_t num_vertex_elements;
  VertexElement *vertex_elements;
  uint32_t num_parts;
  physis_Part *parts;
};

struct physis_MDL {
  MDL *p_ptr;
  uint32_t num_lod;
  physis_LOD *lods;
  uint32_t num_affected_bones;
  const char **affected_bone_names;
  uint32_t num_material_names;
  const char **material_names;
};

struct NewShapeValue {
  uint32_t base_index;
  Vertex replacing_vertex;
};

struct physis_SupportedTribes {
  Tribe subraces[2];
};

struct physis_EXL {
  int32_t version;
  int32_t entry_count;
  const char *const *entry_keys;
  const int32_t *entry_values;
};

struct physis_ColumnData {
  enum class Tag {
    String,
    Bool,
    Int8,
    UInt8,
    Int16,
    UInt16,
    Int32,
    UInt32,
    Float32,
    Int64,
    UInt64,
  };

  struct String_Body {
    const char *_0;
  };

  struct Bool_Body {
    bool _0;
  };

  struct Int8_Body {
    int8_t _0;
  };

  struct UInt8_Body {
    uint8_t _0;
  };

  struct Int16_Body {
    int16_t _0;
  };

  struct UInt16_Body {
    uint16_t _0;
  };

  struct Int32_Body {
    int32_t _0;
  };

  struct UInt32_Body {
    uint32_t _0;
  };

  struct Float32_Body {
    float _0;
  };

  struct Int64_Body {
    int64_t _0;
  };

  struct UInt64_Body {
    uint64_t _0;
  };

  Tag tag;
  union {
    String_Body string;
    Bool_Body bool_;
    Int8_Body int8;
    UInt8_Body u_int8;
    Int16_Body int16;
    UInt16_Body u_int16;
    Int32_Body int32;
    UInt32_Body u_int32;
    Float32_Body float32;
    Int64_Body int64;
    UInt64_Body u_int64;
  };
};

struct physis_ExcelRow {
  physis_ColumnData *column_data;
};

struct physis_ExcelRows {
  physis_ExcelRow *row_data;
  unsigned int row_count;
};

struct physis_Bone {
  uint32_t index;
  const char *name;
  physis_Bone *parent_bone;
  uint32_t parent_index;
  float position[3];
  float rotation[4];
  float scale[3];
};

struct physis_Skeleton {
  uint32_t num_bones;
  physis_Bone *bones;
  physis_Bone *root_bone;
};

struct physis_Texture {
  TextureType texture_type;
  uint32_t width;
  uint32_t height;
  uint32_t depth;
  uint32_t rgba_size;
  uint8_t *rgba;
};

struct ShaderKey {
  uint32_t category;
  uint32_t value;
};

struct Constant {
  uint32_t id;
  uint32_t num_values;
  float values[4];
};

struct Sampler {
  /// This is a CRC hash, it can be calculated via ShaderPackage::crc
  uint32_t texture_usage;
  uint32_t flags;
  uint8_t texture_index;
  uint8_t unknown1;
  uint8_t unknown2;
  uint8_t unknown3;
};

struct LegacyColorTableRow {
  float diffuse_color[3];
  float specular_strength;
  float specular_color[3];
  float gloss_strength;
  float emissive_color[3];
  uint16_t tile_set;
  float material_repeat[2];
  float material_skew[2];
};

struct physis_LegacyColorTable {
  uint32_t num_rows;
  LegacyColorTableRow *rows;
};

struct DawntrailColorTableRow {
  float diffuse_color[3];
  float unknown1;
  float specular_color[3];
  float unknown2;
  float emissive_color[3];
  float unknown3;
  float sheen_rate;
  float sheen_tint;
  float sheen_aperture;
  float unknown4;
  float roughness;
  float unknown5;
  float metalness;
  float anisotropy;
  float unknown6;
  float sphere_mask;
  float unknown7;
  float unknown8;
  uint16_t shader_index;
  uint16_t tile_set;
  float tile_alpha;
  uint16_t sphere_index;
  float material_repeat[2];
  float material_skew[2];
};

struct physis_DawntrailColorTable {
  uint32_t num_rows;
  DawntrailColorTableRow *rows;
};

struct physis_Material {
  const char *shpk_name;
  uint32_t num_textures;
  const char **textures;
  uint32_t num_shader_keys;
  ShaderKey *shader_keys;
  uint32_t num_constants;
  Constant *constants;
  uint32_t num_samplers;
  Sampler *samplers;
  physis_LegacyColorTable legacy_color_table;
  physis_DawntrailColorTable dawntrail_color_table;
};

struct physis_ShaderParameter {
  uint16_t slot;
  const char *name;
};

struct physis_Shader {
  int32_t len;
  uint8_t *bytecode;
  int32_t num_scalar_parameters;
  physis_ShaderParameter *scalar_parameters;
  int32_t num_resource_parameters;
  physis_ShaderParameter *resource_parameters;
};

struct Key {
  uint32_t id;
  uint32_t default_value;
};

struct MaterialParameter {
  uint32_t id;
  uint16_t byte_offset;
  uint16_t byte_size;
};

struct physis_SHPK {
  ShaderPackage *p_ptr;
  int32_t num_vertex_shaders;
  physis_Shader *vertex_shaders;
  int32_t num_pixel_shaders;
  physis_Shader *pixel_shaders;
  int32_t num_system_keys;
  Key *system_keys;
  int32_t num_scene_keys;
  Key *scene_keys;
  int32_t num_material_keys;
  Key *material_keys;
  uint32_t sub_view_key1_default;
  uint32_t sub_view_key2_default;
  uint32_t material_parameters_size;
  uint32_t num_material_parameters;
  MaterialParameter *material_parameters;
};

struct Pass {
  uint32_t id;
  uint32_t vertex_shader;
  uint32_t pixel_shader;
};

struct physis_SHPKNode {
  uint32_t selector;
  uint32_t pass_count;
  uint8_t pass_indices[16];
  uint32_t system_key_count;
  uint32_t *system_keys;
  uint32_t scene_key_count;
  uint32_t *scene_keys;
  uint32_t material_key_count;
  uint32_t *material_keys;
  uint32_t subview_key_count;
  uint32_t *subview_keys;
  Pass *passes;
};

struct physis_CMP {
  CMP *p_ptr;
};

struct RacialScalingParameters {
  /// Male minimum height
  float male_min_size;
  /// Male maximum height
  float male_max_size;
  /// Male minimum tail size
  float male_min_tail;
  /// Male maximum tail size
  float male_max_tail;
  /// Female minimum height
  float female_min_size;
  /// Female maximum height
  float female_max_size;
  /// Female minimum tail size
  float female_min_tail;
  /// Female maximum tail size
  float female_max_tail;
  /// Minimum bust size on the X-axis
  float bust_min_x;
  /// Minimum bust size on the Y-axis
  float bust_min_y;
  /// Minimum bust size on the Z-axis
  float bust_min_z;
  /// Maximum bust size on the X-axis
  float bust_max_x;
  /// Maximum bust size on the Y-axis
  float bust_max_y;
  /// Maximum bust size on the Z-axis
  float bust_max_z;
};

struct CustomizeData {
  /// The race of the character.
  Race race;
  /// The gender of the character.
  Gender gender;
  /// The age of the character. Normal = 1, Old = 3, Young = 4.
  uint8_t age;
  /// The height of the character from 0 to 100.
  uint8_t height;
  /// The character's tribe.
  Tribe tribe;
  /// The character's selected face.
  uint8_t face;
  /// The character's selected hair.
  uint8_t hair;
  /// If hair highlights are enabled for this character.
  bool enable_highlights;
  /// The character's skin tone.
  uint8_t skin_tone;
  /// The character's right eye color.
  uint8_t right_eye_color;
  /// The character's hair color.
  uint8_t hair_tone;
  /// The color of the hair highlights.
  uint8_t highlights;
  /// The selected facial features.
  uint8_t facial_features;
  /// The color of the character's facial features.'
  uint8_t facial_feature_color;
  /// The character's selected eyebrows.
  uint8_t eyebrows;
  /// The character's left eye color.
  uint8_t left_eye_color;
  /// The character's selected eyes.
  /// If the character has selected "Small Iris" then it adds 128 to this.
  uint8_t eyes;
  /// The character's selected nose.
  uint8_t nose;
  /// The character's selected jaw.
  uint8_t jaw;
  /// The character's selected mouth.
  uint8_t mouth;
  /// The character's selected pattern.
  uint8_t lips_tone_fur_pattern;
  /// Depending on the race, it's either the ear size, muscle size, or tail size.
  uint8_t race_feature_size;
  /// Depending on the race, it's the selected tail or ears.
  uint8_t race_feature_type;
  /// The size of the character's bust from 0 to 100.
  uint8_t bust;
  /// The character's choice of face paint.
  uint8_t face_paint;
  /// The color of the face paint.
  /// If the character has selected "Light" then it adds 128 to this.
  uint8_t face_paint_color;
  /// The character's chosen voice.
  uint8_t voice;
};

struct physis_CharacterData {
  uint32_t version;
  CustomizeData customize;
  uint32_t timestamp;
  const char *comment;
};

struct physis_ConfigFile {
  ConfigFile *p_ptr;
};

struct physis_PBD {
  PreBoneDeformer *p_ptr;
};

struct physis_PreBoneDeformBone {
  const char *name;
  float deform[12];
};

struct physis_PreBoneDeformMatrices {
  int32_t num_bones;
  physis_PreBoneDeformBone *bones;
};

struct physis_PlateModel {
  float position[2];
  const char *filename;
};

struct physis_Terrain {
  int32_t num_plates;
  physis_PlateModel *plates;
};

struct physis_Dictionary {
  int32_t num_words;
  const char *const *words;
};

struct physis_IndexEntries {
  uint32_t num_entries;
  const uint32_t *dir_entries;
  const uint32_t *filename_entries;
};

/// An existing install location on disk
struct physis_ExistingGameDirectory {
  /// The application where this installation was from
  ExistingInstallType install_type;
  /// The path to the "main folder" where "game" and "boot" sits
  const char *path;
  /// The latest game/expansion version in this directory
  const char *version;
};

struct physis_ExistingGameDirectories {
  uint32_t count;
  physis_ExistingGameDirectory *entries;
};

/// An existing user directory
struct physis_ExistingUserDirectory {
  /// The application where this directory was from
  ExistingInstallType install_type;
  /// The path to the user folder
  const char *path;
};

struct physis_ExistingUserDirectories {
  uint32_t count;
  physis_ExistingUserDirectory *entries;
};

struct physis_SCHD {
  Schd *p_ptr;
  ShaderStage shader_stage;
};

struct physis_PatchEntry {
  const char *url;
  const char *version;
  uint64_t hash_count;
  const char *const *hashes;
  int64_t hash_block_size;
  int64_t length;
  int64_t size_on_disk;
};

struct physis_PatchList {
  uint64_t patch_length;
  int32_t num_entries;
  const physis_PatchEntry *entries;
  int64_t total_size_downloaded;
};

struct Transformation {
  float translation[3];
  float rotation[3];
  float scale[3];
};

struct physis_BGInstanceObject {
  const char *asset_path;
};

struct physis_LayerEntry {
  enum class Tag {
    None,
    BG,
  };

  struct BG_Body {
    physis_BGInstanceObject _0;
  };

  Tag tag;
  union {
    BG_Body bg;
  };
};

struct physis_InstanceObject {
  uint32_t instance_id;
  const char *name;
  Transformation transform;
  physis_LayerEntry data;
};

struct physis_Layer {
  const physis_InstanceObject *objects;
  uint32_t num_objects;
  const char *name;
  uint32_t id;
};

struct physis_LayerChunk {
  const physis_Layer *layers;
  uint32_t num_layers;
};

struct physis_LayerGroup {
  const physis_LayerChunk *chunks;
  uint32_t num_chunks;
};

extern "C" {

const char *physis_get_physis_version();

const char *physis_get_libphysis_version();

physis_Buffer physis_read_file(const char *path);

void physis_free_file(const physis_Buffer *buffer);

/// Checks if the file at `path` exists.
bool physis_gamedata_exists(GameData *game_data, const char *path);

void physis_gamedata_free_repositories(physis_Repositories repositories);

/// Extracts the raw game file from `path`, and puts it in `data` with `size` length. If the path was not found,
/// `size` is 0 and `data` is NULL.
physis_Buffer physis_gamedata_extract_file(GameData *game_data,
                                           const char *path);

uint64_t physis_gamedata_find_offset(GameData *game_data, const char *path);

void physis_gamedata_free_sheet_header(physis_EXH*);

/// Initializes a new GameData structure.
GameData *physis_gamedata_initialize(const char *path);

void physis_gamedata_free(GameData *game_data);

physis_Repositories physis_gamedata_get_repositories(const GameData *game_data);

physis_EXD physis_gamedata_read_excel_sheet(GameData *game_data,
                                            const char *name,
                                            const physis_EXH *exh,
                                            Language language,
                                            unsigned int page);

const char *physis_gamedata_get_exd_filename(const char *name,
                                             const physis_EXH *exh,
                                             Language language,
                                             unsigned int page);

void physis_gamedata_free_sheet(physis_EXD exd);

bool physis_gamedata_apply_patch(const GameData *gamedata, const char *path);

physis_SheetNames physis_gamedata_get_all_sheet_names(GameData *game_data);

const char *physis_bootdata_get_version(const BootData *boot_data);

/// Initializes a new BootData structure.
BootData *physis_bootdata_initialize(const char *path);

void physis_bootdata_free(BootData *boot_data);

bool physis_bootdata_apply_patch(const BootData *bootdata, const char *path);

physis_MDL physis_mdl_parse(physis_Buffer buffer);

physis_Buffer physis_mdl_write(const physis_MDL *mdl);

void physis_mdl_replace_vertices(physis_MDL *mdl,
                                 uint32_t lod_index,
                                 uint32_t part_index,
                                 uint32_t num_vertices,
                                 const Vertex *vertices_ptr,
                                 uint32_t num_indices,
                                 const uint16_t *indices_ptr,
                                 uint32_t num_submeshes,
                                 const SubMesh *submeshes_ptr);

void physis_mdl_remove_shape_meshes(physis_MDL *mdl);

void physis_mdl_add_shape_mesh(physis_MDL *mdl,
                               uint32_t lod_index,
                               uint32_t shape_index,
                               uint32_t shape_mesh_index,
                               uint32_t part_index,
                               uint32_t num_shape_values,
                               const NewShapeValue *shape_values);

uintptr_t physis_get_vertex_type_size(VertexType vertex_type);

void physis_mdl_free(const physis_MDL *mdl);

int32_t physis_get_race_code(Race race, Tribe tribe, Gender gender);

physis_SupportedTribes physis_get_supported_tribes(Race race);

physis_EXL physis_gamedata_read_excel_list(physis_Buffer buffer);

Slot physis_slot_from_id(int32_t slot_id);

const char *physis_get_slot_name(Slot slot);

const char *physis_build_equipment_path(int32_t model_id,
                                        Race race,
                                        Tribe tribe,
                                        Gender gender,
                                        Slot slot);

const char *physis_build_character_path(CharacterCategory category,
                                        int32_t body_ver,
                                        Race race,
                                        Tribe tribe,
                                        Gender gender);

const char *physis_build_skin_material_path(int32_t race_code,
                                            int32_t body_code,
                                            const char *material_name);

const char *physis_build_gear_material_path(int32_t gear_id,
                                            int32_t gear_version,
                                            const char *material_name);

const char *physis_build_face_material_path(int32_t race_code,
                                            int32_t face_code,
                                            const char *material_name);

const char *physis_build_hair_material_path(int32_t race_code,
                                            int32_t hair_code,
                                            const char *material_name);

const char *physis_build_ear_material_path(int32_t race_code,
                                           int32_t ear_code,
                                           const char *material_name);

const char *physis_build_tail_material_path(int32_t race_code,
                                            int32_t tail_code,
                                            const char *material_name);

Blowfish *physis_blowfish_initialize(uint8_t *key, unsigned int key_size);

void physis_blowfish_free(Blowfish *blowfish);

bool physis_blowfish_encrypt(const Blowfish *blowfish,
                             uint8_t *in_data,
                             unsigned int in_data_size,
                             uint8_t **out_data,
                             uint32_t *out_data_size);

bool physis_blowfish_decrypt(const Blowfish *blowfish,
                             uint8_t *in_data,
                             unsigned int in_data_size,
                             uint8_t **out_data,
                             uint32_t *out_data_size);

SteamTicketBlowfish *miscel_steamticket_blowfish_initialize(uint8_t *key, unsigned int key_size);

void miscel_steamticket_physis_blowfish_free(SteamTicketBlowfish *blowfish);

void miscel_steamticket_blowfish_encrypt(const SteamTicketBlowfish *blowfish,
                                         uint8_t *in_data,
                                         unsigned int in_data_size);

physis_EXH *physis_parse_excel_sheet_header(physis_Buffer buffer);

physis_ExcelRows physis_exd_read_row(const physis_EXD *exd, uint32_t id);

void physis_exd_free_rows(const physis_EXD *exd, const physis_ExcelRows *rows);

physis_Skeleton physis_parse_skeleton(physis_Buffer buffer);

physis_Texture physis_texture_parse(physis_Buffer buffer);

physis_Material physis_material_parse(physis_Buffer buffer);

physis_SHPK physis_parse_shpk(physis_Buffer buffer);

physis_SHPKNode physis_shpk_get_node(const physis_SHPK *shpk, uint32_t key);

uint32_t physis_shpk_build_selector_from_all_keys(const uint32_t *system_keys,
                                                  uint32_t system_key_count,
                                                  const uint32_t *scene_keys,
                                                  uint32_t scene_key_count,
                                                  const uint32_t *material_keys,
                                                  uint32_t material_key_count,
                                                  const uint32_t *subview_keys,
                                                  uint32_t subview_key_count);

uint32_t physis_shpk_crc(const char *name);

physis_CMP physis_cmp_parse(physis_Buffer buffer);

RacialScalingParameters physis_cmp_get_racial_scaling_parameters(physis_CMP cmp, Race, Tribe tribe);

physis_CharacterData physis_chardat_parse(physis_Buffer buffer);

physis_ConfigFile physis_cfg_parse(physis_Buffer buffer);

void physis_cfg_set_value(physis_ConfigFile cfg, const char *key, const char *value);

physis_Buffer physis_cfg_write(physis_ConfigFile cfg);

physis_PBD physis_parse_pbd(physis_Buffer buffer);

physis_PreBoneDeformMatrices physis_pbd_get_deform_matrix(physis_PBD pbd,
                                                          uint16_t from_body_id,
                                                          uint16_t to_body_id);

physis_Terrain physis_parse_tera(physis_Buffer buffer);

physis_Dictionary physis_parse_dictionary(physis_Buffer buffer);

physis_IndexEntries physis_index_parse(const char *path);

uint32_t physis_generate_partial_hash(const char *name);

uint64_t physis_calculate_hash(const char *index_file_path, const char *path);

physis_ExistingGameDirectories physis_find_existing_game_dirs();

physis_ExistingUserDirectories physis_find_existing_user_dirs();

physis_SCHD physis_parse_schd(physis_Buffer _buffer);

const char *physis_extract_frontier_url(const char *launcher_path);

physis_PatchList physis_parse_patchlist(PatchListType patch_type, const char *encoded);

physis_LayerGroup physis_layergroup_read(physis_Buffer buffer);

}  // extern "C"
